/* wptDNSKeys.cpp - Support for retrieving keys via DNS
 *	Copyright (C) 2006 Timo Schulz
 *
 * This file is part of WinPT.
 *
 * WinPT is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * WinPT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with WinPT; if not, write to the Free Software Foundation, 
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <windows.h>
#include <windns.h>

#include "wptDNSKeys.h"


/* typedef for the function signature. */
typedef DNS_STATUS (*dns_query_fnc) (LPSTR, WORD, DWORD, PIP4_ARRAY,
				    PDNS_RECORD*, PVOID *);

typedef void (*dns_record_list_free_fnc) (PDNS_RECORD, DNS_FREE_TYPE);

/* function pointer. */
static dns_query_fnc dns_query = NULL;
static dns_record_list_free_fnc dns_record_free = NULL;

/* hinstance handle to the DLL. */
static HINSTANCE dns_api = NULL;

/* 1 if the DNS api is not available. */
static int dns_failed = 0;



/* Initialize the DNS sub system. We do this via dynamic loading
   because older NT/9X systems do not have this API. */
static int
dns_init (void)
{
    if (dns_query)
	return 0;
    if (dns_failed)
	return -1;
    dns_api = LoadLibrary ("dnsapi");
    if (!dns_api) {
	dns_failed = 1;
	return -1;
    }
    dns_query = (dns_query_fnc)GetProcAddress (dns_api, "DnsQuery_A");
    if (!dns_query) {
	dns_failed = 1;
	return -1;
    }
    dns_record_free = (dns_record_list_free_fnc)GetProcAddress (dns_api, "DnsRecordListFree");
    if (!dns_record_free) {
	dns_failed = 1;
	return -1;
    }
    return 0;
}


/* Cleanup static structs. */
void
dns_cleanup (void)
{
    if (dns_api)
	FreeLibrary (dns_api);
    dns_api = NULL;
    dns_failed = 0;
}




/* build a DNS name for the PKA lookup. */
static char*
email_get_pka_addr (const char *uid)
{
    const char *fmt = "._pka.";
    char *bo;
    char *pka;
    int pos=0;

    if ((bo=strchr (uid, '<')) && strchr (uid, '>'))
	uid += (bo-uid+1);
    if (!strchr (uid, '@'))
	return NULL;

    pka = (char*)calloc (1, strlen (uid)+strlen (fmt)+1);
    while (uid && *uid != '@')
	pka[pos++] = *uid++;
    uid++;
    strcat (pka, fmt); pos += strlen (fmt);
    while (uid && *uid && *uid != '>')
	pka[pos++] = *uid++;
    return pka;
}


/* Convert the returned data from the PKA (txt) record. */
pka_info_t
parse_pka_data (const char *data)
{
    enum pka_col_t { COL_VER=1, COL_FPR, COL_URI };
    pka_info_t pka;
    char *p;
    int pos = 1;

    if (strncmp (data, "v=pka1;", 8))
	return NULL;
    pka = (pka_info_t)calloc (1, sizeof *pka);
    p = strtok ((char*)data, ";");
    while (p != NULL) {
	switch (pos) {
	case COL_VER: 
	    pka->ver = 1; 
	    break;

	case COL_FPR: 
	    pka->fpr = strdup (p+strlen ("fpr=")); 
	    break;

	case COL_URI:  /* optional */
	    pka->uri = strdup (p+strlen ("uri=")); 
	    break;

	default:
	    break;
	}
	pos++;
    }
    if (pos != 3) {
	dns_free_pka_record (pka);
	pka = NULL;
    }
	
    return pka;
}


/* Retrieve a PKA record from the DNS.
   @userid is used to extract the email address. */
extern "C" int
dns_get_pka_record (const char *userid, pka_info_t *r_pka)
{
    char *addr;
    DNS_STATUS err;
    DNS_RECORD *rec;

    *r_pka = NULL;
    if (dns_init ())
	return -1;
    addr = email_get_pka_addr (userid);
    if (!addr)
	return -1;
    err = dns_query (addr, DNS_TYPE_TEXT, 0, NULL, &rec, NULL);
    if (err) {
	free (addr);
	return -1;
    }
    *r_pka = parse_pka_data (rec->Data.Txt.pStringArray[0]);

    dns_record_free (rec, DnsFreeRecordList);
    free (addr);
    return 0;
}


/* Release the memory of the @pka structure. */
extern "C" void
dns_free_pka_record (pka_info_t pka)
{
    if (pka->fpr)
	free (pka->fpr);
    pka->fpr = NULL;
    if (pka->uri)
	free (pka->uri);
    pka->uri = NULL;
    free (pka);
}
