/* wptCommonDlg.cpp
 *	Copyright (C) 2004, 2005, 2006 Timo Schulz
 *
 * This file is part of WinPT.
 *
 * WinPT is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * WinPT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with WinPT; if not, write to the Free Software Foundation, 
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <windows.h>
#include <ctype.h>

#include "resource.h"
#include "wptTypes.h"
#include "wptW32API.h"
#include "wptVersion.h"
#include "wptGPG.h"
#include "wptCommonCtl.h"
#include "wptContext.h"
#include "wptRegistry.h"


/* check if the given buf contains a valid URL.
   for the moment we hardcode HTTP urls. */
static int
check_URL (const char *buf)
{
    size_t i;
    const char *allowed = "-./#_:~&?=";

    if (strncmp (buf, "http", 4))
	return -1;
    if (!strchr (buf, ':'))
	return -1;
    if (!strstr (buf, "//"))
	return -1;
    if (strstr (buf, " "))
	return -1;
    buf += 7;
    for (i=0; i < strlen (buf); i++) {
	if (isalpha (buf[i]) || isdigit (buf[i]))
	    continue;
	if (!strchr (allowed, buf[i]))
	    return -1;
    }

    return 0;
}


static BOOL CALLBACK
http_dlg_proc (HWND dlg, UINT msg, WPARAM wparam, LPARAM lparam)
{
    static struct URL_ctx_s * ctx;
    int n;

    switch (msg) {
    case WM_INITDIALOG:
	ctx = (struct URL_ctx_s*)lparam;
	if (!ctx)
	    BUG (0);
	if (ctx->desc != NULL)
	    SetWindowText (dlg, ctx->desc);
	if (ctx->title != NULL)
	    SetDlgItemText (dlg, IDC_HTTP_TITLE, ctx->title);
	SetDlgItemText (dlg, IDCANCEL, _("&Cancel"));
	SetForegroundWindow (dlg);
	break;

    case WM_COMMAND:
	switch (LOWORD (wparam)) {
	case IDOK:
	    n = GetDlgItemText (dlg, IDC_HTTP_URL, ctx->url, DIM (ctx->url)-1);
	    if (ctx->check && (!n || check_URL (ctx->url))) {
		msg_box (dlg, _("Please enter a valid URL."), "HTTP", MB_ERR);
		return FALSE;
	    }
	    ctx->cancel = 0;
	    EndDialog (dlg, TRUE);
	    break;

	case IDCANCEL:
	    ctx->cancel = 1;
	    EndDialog (dlg, FALSE);
	    break;
	}
    }

    return FALSE;
}


/* Common dialog to request a URL from the user. 
   Return value is the URL context. */
void*
get_http_file_dlg (HWND root)
{
    struct URL_ctx_s *ctx;

    ctx = new URL_ctx_s;
    if (!ctx)
	BUG (0);
    memset (ctx, 0, sizeof *ctx);
    ctx->check = 1;
    ctx->desc = _("HTTP Key Import");
    ctx->title = _("Enter URL to retrieve the public key");
    DialogBoxParam (glob_hinst, (LPCTSTR)IDD_WINPT_HTTP, root,
		    http_dlg_proc, (LPARAM)ctx);
    return ctx;
}

/* Wrapper to request a preferred keyserver from the user.
   Return value is the URL context. */
void*
get_keyserver_URL_dlg (HWND root)
{
    struct URL_ctx_s *ctx;

    ctx = new URL_ctx_s;
    if (!ctx)
	BUG (0);
    memset (ctx, 0, sizeof *ctx);
    ctx->check = 1;
    ctx->desc = _("Key Edit");
    ctx->title = _("Enter preferred keyserver URL"); 
    DialogBoxParam (glob_hinst, (LPCTSTR)IDD_WINPT_HTTP, root,
		    http_dlg_proc, (LPARAM)ctx);
    return ctx;
}


/* Generic text input dialog. */
char*
get_input_dialog (HWND root, const char *title, const char *desc)
{
    struct URL_ctx_s ctx;

    memset (&ctx, 0, sizeof (ctx));
    ctx.title = title;
    ctx.desc = desc;
    DialogBoxParam (glob_hinst, (LPCTSTR)IDD_WINPT_HTTP, root,
		    http_dlg_proc, (LPARAM)&ctx);
    if (ctx.cancel || strlen (ctx.url) == 0)
	return NULL;
    return m_strdup (ctx.url);
}


/* Initialize the language list based on the .mo files
   found in the dirctory @modpath. */
static int
nls_load_langlist (HWND dlg, const char *modpath)
{
    WIN32_FIND_DATA fnd;
    HWND cb;
    HANDLE hd;
    char *pattern = NULL;    
    int i;

    cb = GetDlgItem (dlg, IDC_LANGUAGE_LIST);
    combox_add_string (cb, lang_list[0].name);
    SendMessage (cb, CB_SETCURSEL, 0, 0);
    
    memset (&fnd, 0, sizeof (fnd));
    pattern = make_filename (modpath, "*.mo", NULL);
    hd = FindFirstFile (pattern, &fnd);
    if (hd == INVALID_HANDLE_VALUE) {
	free_if_alloc (pattern);
	return -1;
    }

    do {
	for (i=0; lang_list[i].id != NULL; i++) {
	    if (!strnicmp (fnd.cFileName, lang_list[i].id, 2))
		combox_add_string (cb, lang_list[i].name);
	}
	memset (&fnd, 0, sizeof (fnd));
    } while (FindNextFile (hd, &fnd) != FALSE);
    FindClose (hd);
    free_if_alloc (pattern);
    return 0;
}


/* Set the default language for the program.
   The $lang.mo file is copied either to the selected
   folder or the directory specified by @modpath.
   The output name of the file is winpt.mo (example: de.mo -> winpt.mo). */
static int
nls_set_language (HWND dlg, const char *modpath)
{
    HWND cb = GetDlgItem (dlg, IDC_LANGUAGE_LIST);
    int pos = SendMessage (cb, CB_GETCURSEL, 0, 0);
    const char *folder;
    char *src, *dst;

    /* default langauge is English. */
    if (pos == 0 || pos == CB_ERR)
	return 0;

    folder = get_folder_dlg (dlg, _("Choose Locale Directory"), modpath);
    if (!folder)
	return 0;

    src = make_filename (modpath, lang_list[pos].id, "mo");
    dst = make_filename (folder, "winpt", "mo");

    if (!CopyFile (src, dst, FALSE)) {
	msg_box (dlg, _("Could not create winpt.mo file"), 
		 _("WinPT Error"), MB_ERR);
	pos = -1;
    }
    else
	set_reg_entry_mo (folder);
    free_if_alloc (src);
    free_if_alloc (dst);
    return pos;
}


static BOOL CALLBACK
nls_dlg_proc (HWND dlg, UINT msg, WPARAM wparam, LPARAM lparam)
{
    static char modpath[MAX_PATH+1];
    int langid = 0, n;

    switch (msg) {
    case WM_INITDIALOG:
	if ((n=GetModuleFileName (NULL, modpath, MAX_PATH-1)) > 0) {
	    while (n-- > 0) {
		if (modpath[n] == '\\') {
		    modpath[n] = 0;
		    break;
		}
	    }
	}
	n = nls_load_langlist (dlg, modpath);
	if (n == -1)
	    EndDialog (dlg, -1);
	SetWindowText (dlg, _("Native Language Support"));
	SetDlgItemText (dlg, IDC_LANGUAGE_INFO, _("Please select a language"));
	SetDlgItemText (dlg, IDOK, _("&OK"));
	SetDlgItemText (dlg, IDCANCEL, _("&Cancel"));
	center_window (dlg, NULL);
	SetForegroundWindow (dlg);
	break;

    case WM_COMMAND:
	switch (LOWORD (wparam)) {
	case IDOK:
	    langid = nls_set_language (dlg, modpath);
	    EndDialog (dlg, langid);
	    break;

	case IDCANCEL:
	    EndDialog (dlg, -1);
	    break;
	}
	break;
    }

    return FALSE;
}


int
select_language (void)
{
    DialogBoxParam (glob_hinst, (LPCTSTR)IDD_WINPT_LANGUAGE, GetDesktopWindow (),
		    nls_dlg_proc, 0);
    return 0;
}

