/* wptFileCBS.cpp - Customized I/O callbacks for GPGME
 *	Copyright (C) 2005, 2006 Timo Schulz
 *	Copyright (C) 2005 g10 Code GmbH
 *
 * This file is part of WinPT.
 *
 * WinPT is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either version 2 
 * of the License, or (at your option) any later version.
 *  
 * WinPT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with WinPT; if not, write to the Free Software Foundation, 
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <windows.h>
#include <stdio.h>
#include <errno.h>

#include "wptListView.h"
#include "wptGPG.h"
#include "wptFileManager.h"
#include "wptErrors.h"
#include "wptTypes.h"
#include "wptW32API.h"

void progress_callback (void *opaque, const char *what, int type, 
			int off, int max);


/* Predefined read callback. */
static long
read_cb (void *handle, void *buffer, size_t size)
{
    file_data_t cb = (file_data_t)handle;
    struct progress_filter_s *pfx = (struct progress_filter_s *)cb->cb_value;
    DWORD nread = 0;

    if (cb->error)
	return -1;

    if (!ReadFile (cb->handle, buffer, size, &nread, NULL)) {	
	cb->error = (int)GetLastError ();
	log_debug ("read_cb: ReadFile() error=%d\r\n", cb->error);
	return -1;
    }

    /* XXX: there is a sync problem with the progress dialog. */
    if (pfx)
	progress_callback (pfx, NULL, 0, cb->off, cb->size);
    cb->off += nread;
    return (long)nread;
}


/* Predefined write callback. */
static long
write_cb (void *handle, const void *buffer, size_t size)
{
    file_data_t cb = (file_data_t)handle;
    DWORD nwritten;

    if (cb->error)
	return -1;

    if (!cb->handle) {
	SECURITY_ATTRIBUTES sec_attr;

	memset (&sec_attr, 0, sizeof (sec_attr));
	sec_attr.bInheritHandle = FALSE;
	sec_attr.nLength = sizeof (sec_attr);
	cb->handle = CreateFile (cb->name, GENERIC_WRITE, FILE_SHARE_WRITE,
				 &sec_attr, CREATE_ALWAYS, 0, NULL);
	if (cb->handle == INVALID_HANDLE_VALUE) {
	    cb->error = (int)GetLastError ();
	    log_debug ("write_cb: CreateFile() error=%d\r\n", cb->error);
	    return -1;
	}
    }

    if (!WriteFile (cb->handle, buffer, size, &nwritten, NULL)) {
	cb->error = (int)GetLastError ();
	log_debug ("write_cb: WriteFile() error=%d\r\n", cb->error);
	return -1;
    }
    return (long)nwritten;
}


/* Create a new data -> file association with a static callback.
   @fname is the file which is associated to the object.
   @r_cb is the context which holds all information.
   @for_read is 1 if the file is opened for read only.
   Return value: 0 on success. */
gpgme_error_t
gpg_file_data_new (const char *fname, int flags, file_data_t *r_cb)
		      
{
    gpgme_error_t err;
    file_data_t cb;
    HANDLE fd = NULL;
    SECURITY_ATTRIBUTES sec_attr;

    memset (&sec_attr, 0, sizeof (sec_attr));
    sec_attr.bInheritHandle = FALSE;
    sec_attr.nLength = sizeof (sec_attr);
    if (flags & F_DATA_READ) {
	fd = CreateFile (fname, GENERIC_READ, FILE_SHARE_READ,
			 &sec_attr, OPEN_EXISTING, 0, NULL);
	if (fd == INVALID_HANDLE_VALUE) {
	    log_debug ("gpg_file_data_new: CreateFile() error=%d\r\n", 
		       (int)GetLastError ());
	    return gpgme_err_code_from_errno (ENOENT);
	}
    }
    cb = new file_data_s;
    if (!cb)
	BUG (NULL);
    memset (cb, 0, sizeof *cb);
    cb->name = m_strdup (fname);
    cb->cbs.read = read_cb;
    cb->cbs.write = write_cb;
    if (flags & F_DATA_READ) {
	cb->handle = fd;
	cb->size = GetFileSize (fd, NULL);
	cb->off = 0;
    }

    err = gpgme_data_new_from_cbs  (&cb->dat, &cb->cbs, cb);
    if (err) {
	CloseHandle (fd);
	free_if_alloc (cb);
	return err;
    }

    *r_cb = cb;
    return err;
}


/* Activate the progress callback for the given object @ctx. */
void
gpg_file_data_set_cb (file_data_t ctx, struct progress_filter_s *pfx)
{
    ctx->cb_value = (void*)pfx;
}


/* Release the context in @cb. Close all internal handles if possible. */
void
gpg_file_data_release (file_data_t cb)
{
    if (!cb)
	return;
    if (cb->handle) {
	CloseHandle (cb->handle);
	cb->handle = NULL;
    }
    if (cb->dat)
	gpgme_data_release (cb->dat);
    free_if_alloc (cb->name);
    free_if_alloc (cb);
}
