/* wptImportList.cpp - Key import list
 *	Copyright (C) 2001-2006 Timo Schulz
 *	Copyright (C) 2005 g10 Code GmbH
 *
 * This file is part of WinPT.
 *
 * WinPT is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either version 2 
 * of the License, or (at your option) any later version.
 *  
 * WinPT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with WinPT; if not, write to the Free Software Foundation, 
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <windows.h>
#include <sys/types.h>
#include <ctype.h>

#include "wptTypes.h"
#include "wptGPG.h"
#include "wptCommonCtl.h"
#include "wptKeylist.h"
#include "wptNLS.h"
#include "wptErrors.h"
#include "wptUTF8.h"
#include "wptW32API.h"
#include "wptRegistry.h"
#include "openpgp.h"


/* Symbolic column ids */
enum impl_col_t {
    IMPL_COL_UID	= 0,
    IMPL_COL_SIZE	= 1,
    IMPL_COL_KEYID	= 2,
    IMPL_COL_CREATION	= 3,
    IMPL_COL_KTYPE	= 4
};

/* context to hold information about a key. */
struct import_key_s {
    /* primary userid */
    char *uid;

    /* 1 = secret key is available. */
    int secret;

    /* flags */
    unsigned int disabled:1;
    unsigned int expired:1;
    unsigned int revoked:1;

    /* primary key attributes */
    char keyid[16+1];		
    gpgme_pubkey_algo_t pubkey_algo;
    size_t length;
    time_t timestamp;

    /* secondary key attributes */
    char subkey_keyid[16+1];
    gpgme_pubkey_algo_t subkey_algo;
    size_t subkey_length;
    time_t subkey_timestamp;
};
typedef struct import_key_s *import_key_t;


/* Release the key in @c. */
static void
import_key_release (import_key_t c)
{
    if (!c)
	return;
    safe_free (c->uid);
    safe_free (c);
}


/* Decode the %-encoded userid in @pend and store it as the first
   userid in the key @c. */
static void
decode_userid (char *pend, import_key_t c)
{
    char *uid;

    uid = (char*)calloc (1, strlen (pend) + 1);
    if (!uid)
	BUG (NULL);
    gpg_decode_c_string (pend, &uid, strlen (pend) + 1);
    uid[strlen (uid) -3] = '\0';
    c->uid = utf8_to_native (uid);
    safe_free (uid);
}


/* Parse the output of 'gpg --with-colons <data'. */
static void
parse_colon_key (char *buf, import_key_t c)
{
    const char * s;
    char *p, *pend;
    int field = 0, rectype = 0;
    enum key_t {
	KEY_none      = 0,
	KEY_primary   = 1,
	KEY_secondary = 2,
    };
    
    if (!buf)
        return; /* EOF */    

    if( !strncmp (buf, "pub", 3) || !strncmp (buf, "sec", 3)) {
	rectype = KEY_primary;
	if (*buf == 's')
	    c->secret = 1;
    }
    else if (!strncmp (buf, "sub", 3) || !strncmp (buf, "ssb", 3))
	rectype = KEY_secondary;    
    else if (!strncmp (buf, "uid", 3)) {
	if (!c->uid) {
	    const char *uid = buf+3;
	    while (uid && *uid == ':')
		uid++;
	    decode_userid ((char *)uid, c);
	}
	return;
    }
    else
	return;

    for (p = buf; p; p = pend) {
        field++;
        pend = strchr (p, ':');
        if (pend)
            *pend++ = 0;

        switch (field) {
	case 1:
	    if (rectype != KEY_primary)
		break;
	    for (s = pend; *s && !isdigit (*s); s++) {
		switch (*s) {
		case 'd': c->disabled = 1;break;
		case 'e': c->expired = 1; break;
		case 'r': c->revoked = 1; break;
		}
	    }
	    break;

        case 2:
	    if (rectype == KEY_primary)
		c->length = atoi (pend);
	    else if (rectype == KEY_secondary)
		c->subkey_length = atoi (pend);
            break;

        case 3:
	    if (rectype == KEY_primary)
		c->pubkey_algo = (gpgme_pubkey_algo_t)atoi (pend);
	    else if (rectype == KEY_secondary)
		c->subkey_algo = (gpgme_pubkey_algo_t)atoi (pend);
            break;

        case 4:
	    if (rectype == KEY_primary) {
		strncpy (c->keyid, pend, 16);
		c->keyid[16] = 0;
	    }
	    else if (rectype == KEY_secondary) {
		strncpy (c->subkey_keyid, pend, 16);
		c->subkey_keyid[16] = 0;
	    }
            break;

        case 5:
	    if (rectype == KEY_primary)
		c->timestamp = strtoul (pend, NULL, 10);
	    else if (rectype == KEY_secondary)
		c->subkey_timestamp = strtoul (pend, NULL, 10);
            break;

        case 9:
	    if (rectype == KEY_primary && !c->uid && strlen (pend) > 2) {
		if (!strchr (pend, '[') && !strchr (pend, ']'))
		    decode_userid (pend, c);
	    }
            break;
        }
    }
}

/* Read the next key from data stream @out. The result is stored in @r_key.
   Return value: 0 on success. */
gpgme_error_t
gpg_import_next_key (gpgme_data_t out, char **pending_line, import_key_t *r_key)
{
    import_key_t key;
    int in_cert, got_block = 0;
    char buf[384];
    int n=0;

    if (!r_key)
	return gpg_error (GPG_ERR_INV_ARG);

    key = (import_key_t) calloc (1, sizeof *key);
    if (!key)
	return gpg_error (GPG_ERR_ENOMEM);

    if (*pending_line) {
	parse_colon_key (*pending_line, key);
	safe_free (*pending_line);
	*pending_line = NULL;
	in_cert = 1;
    }
    else
	in_cert = 0;

    while ((n=gpg_data_readline (out, buf, sizeof buf-1)) > 0) {
	if (!strncmp (buf, "pub", 3) || !strncmp (buf, "sec", 3)) {
	    if (in_cert) {
		*pending_line = strdup (buf);
		goto ready;
	    }
	    in_cert = 1;
	    got_block = 1;
	    parse_colon_key (buf, key);
	}
	else if (in_cert) {
	    parse_colon_key (buf, key);
	    got_block = 1;
	}
    }
ready:
    *r_key = key;
    if (got_block && n == 0)
	return 0;
    return n == 0? gpg_error (GPG_ERR_EOF) : 0;
}


/* Return a humand readable key description of @key. */
static char*
key_description (import_key_t key)
{
    gpgme_pubkey_algo_t pkalgo;
    gpgme_pubkey_algo_t subalgo = (gpgme_pubkey_algo_t)0;
    const char *type, *state;
    char *p;
    int n=0;
    
    pkalgo = key->pubkey_algo;
    if (key->subkey_algo)
	subalgo = key->subkey_algo;
    if (key->revoked)
	state = _("Revoked" );
    else if (key->expired)
	state = _("Expired" );
    else
	state = "";
    if (key->secret)
	type = _("secret key");
    else
	type = _("public key");

    n = strlen (state) + strlen (type) + 2*8 + 8;
    p = new char[n+1];
    if (!p)
	BUG (NULL);

    if (!subalgo)
	_snprintf (p, 64, "%s %s %s", state,
		   get_key_pubalgo (pkalgo),type);
    else
	_snprintf (p, 64, "%s %s/%s %s", state,
		    get_key_pubalgo (pkalgo),
		    get_key_pubalgo (subalgo), type);
    return p;
}


/* Add the key @key to the list view control @lv at position @pos. */
static int
implist_add_key (listview_ctrl_t lv, int pos, import_key_t key)
{
    char buf[128], * desc;
    const char *t;
    u32 tt, tt2=0;
	
    if (listview_add_item( lv, " " ))
	return WPTERR_GENERAL;
    t = key->uid;
    if (!t || strlen (t) < 5)
	t = _("Invalid user ID");
    listview_add_sub_item (lv, pos, IMPL_COL_UID, t);

    tt = key->length;
    if (key->subkey_length > 0)
	tt2 = key->subkey_length;
    if (tt && tt2)
	_snprintf (buf, sizeof buf - 1, "%d/%d", tt, tt2);
    else
	_snprintf (buf, sizeof buf-1, "%d", tt);
    listview_add_sub_item (lv, pos, IMPL_COL_SIZE, buf);

    t = key->keyid;
    if (!t || strlen (t) < 8)
	t = "????????????????";
    _snprintf (buf, sizeof buf -1, "0x%s", t+8);
    listview_add_sub_item (lv, pos, IMPL_COL_KEYID, buf);

	
    tt = key->timestamp;
    t = get_key_created (tt);
    if( !t )
	t = "????" "-??" "-??";
    listview_add_sub_item (lv, pos, IMPL_COL_CREATION, (char *)t);

    desc = key_description (key);
    if (desc) {
	listview_add_sub_item (lv, pos, IMPL_COL_KTYPE, desc);
	free_if_alloc (desc);
    }

    return 0;
}


/* Create a list view for list keys. */
int
implist_build (listview_ctrl_t *lv, HWND ctrl)
{
    struct listview_ctrl_s *c;
    struct listview_column_s implist[] = {
	{0, 190, (char *)_("User ID")},	
	{1,  66, (char *)_("Size")},
	{2,  80, (char *)_("Key ID")},
	{3,  72, (char *)_("Creation")},
	{4, 132, (char *)_("Type")},
	{0,   0, NULL}
	};	
    int j;
	
    listview_new (&c, ctrl);
    for (j = 0; implist[j].fieldname != NULL; j++)
	listview_add_column (c, &implist[j]);
    listview_set_ext_style (c);
    *lv = c;
    return 0;
}


static int
check_import_data (const char *fname, int *r_revcert)
{
    FILE *f;
    char buf[512];
    int n;

    f = fopen (fname, "rb");
    if (!f)
	return -1;

    *r_revcert = 0;
    n = fread (buf, 1, 511, f);
    if (n > 0) {
	buf[n] = 0;
	if (strstr (buf, "A revocation certificate should follow"))
	    *r_revcert = 1;
    }

    fclose (f);
    return 0;
}


/* Load the list view @lv with the contents of the file @file.
   @r_revcerts contains the number of revocation certs and
   @r_seckeys the number of imported secret keys. */
int
implist_load (listview_ctrl_t lv, const char *file, 
	      int *r_revcerts, int *r_seckeys)
{
    gpgme_data_t list;
    gpgme_error_t err;    
    import_key_t key;
    int rc = 0, is_tmp=0;
    char *pending_line = NULL;
    char fname[300], *out;
    
    if (!file) {
	err = gpg_data_new_from_clipboard (&list, 0);
	if (err) {
	    msg_box (NULL, gpgme_strerror (err), _("Import"), MB_ERR);
	    return WPTERR_CLIP_OPEN;
	}
	get_temp_name (fname, sizeof (fname)-1, "in_gpg_keys");
	err = gpg_data_release_and_set_file (list, fname);
	if (err) {
	    gpgme_data_release (list);
	    msg_box (NULL, gpgme_strerror (err), _("Import"), MB_ERR);
	    return WPTERR_FILE_CREAT;
	}
	file = fname;
	is_tmp = 1;
    }

    if (check_import_data (file, r_revcerts)) {
	msg_box (NULL, _("It is possible that the ASCII-Armor is damaged\n"
			 "and thus causing a CRC error."), 
		 _("Import"), MB_ERR);
	if (is_tmp)
	    DeleteFile (file);
	return WPTERR_GENERAL;
    }

    err = gpg_import_key_list (file, &out);
    if (err) {
	msg_box (NULL, gpgme_strerror (err), _("Import"), MB_ERR);
	if (is_tmp) 
	    DeleteFile (file);
	return WPTERR_GENERAL;	
    }

    err = gpgme_data_new_from_mem (&list, out, strlen (out), 1);
    free (out);
    if (err) {
	msg_box (NULL, gpgme_strerror (err), _("Import"), MB_ERR);
	if (is_tmp) 
	    DeleteFile (file);
	return WPTERR_GENERAL;
    }

    if (r_seckeys)
	*r_seckeys = 0;
    while (!rc) {
	/* XXX if the key has a direct key signature, the user-id field 
	       in the --with-colons mode is empty! */
	rc = gpg_import_next_key (list, &pending_line, &key);
	if (!rc)
	    rc = implist_add_key (lv, 0, key);
	if (key->secret)
	    (*r_seckeys)++;
	import_key_release (key);
    }

    gpgme_data_release (list);
    if (is_tmp) 
	DeleteFile (file);
    return rc;
}


/* Release the list view @lv. */
void 
implist_delete (listview_ctrl_t lv)
{
    if (lv) {
	listview_release (lv);
    }
}
