/* wptKeyserverSearchDlg.cpp - Keyserver Searching
 *	Copyright (C) 2001-2006 Timo Schulz
 *
 * This file is part of WinPT.
 *
 * WinPT is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * WinPT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with WinPT; if not, write to the Free Software Foundation, 
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <windows.h>
#include <stdio.h>

#include "resource.h"
#include "wptCommonCtl.h"
#include "wptKeyserver.h"
#include "wptErrors.h"
#include "wptTypes.h"
#include "wptGPG.h"
#include "wptDlgs.h"
#include "wptNLS.h"
#include "wptKeylist.h"


/* Default threshold for key items. */
#define MAX_N_KEYS 64


/* Symbolic column IDs. */
enum { KSS_COL_BITS = 0, KSS_COL_ALGO, 
       KSS_COL_KEYID, KSS_COL_DATE, KSS_COL_UID };

int hkp_recv_key2 (HWND dlg, const char *kserver, WORD port,
	           const char *pattern, int proto, char **r_fpr);

/* Initialize dialog. */
static listview_ctrl_t
on_init_dialog (HWND dlg, UINT msg, WPARAM wparam, LPARAM lparam)
{
    struct listview_column_s keysearch[] = {
        {0,  48, (char *)_("Size")},
	{1,  48, (char*)_("Algorithm")},
        {2,  80, (char *)_("Key ID")},
        {3,  72, (char *)_("Creation")},
        {4, 256, (char *)_("User ID")},
        {0,   0, NULL}
    };
    listview_ctrl_t lv = NULL;
    char info[300];
    int i;

    SetDlgItemText (dlg, IDC_HKPSEARCH_RECV, _("&Receive"));
    SetDlgItemText (dlg, IDOK, _("&Cancel"));
    _snprintf (info, DIM (info) - 1,
	       _("Connect to '%s' to search for \"%s\""),
		((keyserver_ctx *)lparam)->name, 
		((keyserver_ctx *)lparam)->pattern);
    SetDlgItemText (dlg, IDC_HKPSEARCH_INFO, info);
    SetWindowText (dlg, _("Keyserver Searching"));
    SetForegroundWindow (dlg);
    listview_new (&lv, GetDlgItem (dlg, IDC_HKPSEARCH_LIST));
    for (i = 0; keysearch[i].width; i++)
        listview_add_column (lv, &keysearch[i]);       
    listview_set_ext_style (lv);
    return lv;
}


/* Begin the key serach and add all results to the list view @lv.
   Return value: 0 on success. */
static int
search_hkp_keys (HWND dlg, keyserver_ctx *ksc, listview_ctrl_t lv)
{
    keyserver_key_s *key;
    const char *uid;
    char bits[32], algo[32], keyid[32];
    int conn_fd = 0;    
    int rc, nkeys = 0;

    rc = kserver_search_begin (ksc->name, ksc->port, ksc->pattern, 
			       &conn_fd, &nkeys);
    if (rc) {
        msg_box (dlg, winpt_strerror (rc), _("Keyserver Search"), MB_ERR);
	return rc;
    }

    if (nkeys > MAX_N_KEYS) {
	rc = log_box (_("Keyserver Search"), MB_INFO_ASK, 
		      _("The search result contains a lot of keys: %d\n\n"
		        "Do you really want to continue?"), nkeys);
	if (rc == IDNO) {
	    kserver_search_end (conn_fd);
	    return -1;
	}
    }

    while (nkeys-- > 0) {
	if (kserver_search_next (conn_fd, &key))
	    break;
	if (!key)
	    continue;

	_snprintf (bits, DIM (bits) - 1, "%d", key->bits);
	_snprintf (algo, DIM (algo)-1, "%s", 
		    get_key_pubalgo ((gpgme_pubkey_algo_t)key->algo));
	_snprintf (keyid, DIM (keyid)-1, "0x%s", key->keyid);
	uid = key->main_uid? key->main_uid->uid : "";
	listview_add_item (lv, " ");
	listview_add_sub_item (lv, 0, KSS_COL_BITS, bits);
	listview_add_sub_item (lv, 0, KSS_COL_ALGO, algo);
	listview_add_sub_item (lv, 0, KSS_COL_KEYID, keyid);
	listview_add_sub_item (lv, 0, KSS_COL_DATE, get_key_created (key->creation));
	listview_add_sub_item (lv, 0, KSS_COL_UID, uid);
	kserver_release_key (key); key=NULL;
    }

    kserver_search_end (conn_fd);
    return 0;
}


/* Fetch all selected keys from the keyserver in @ksc. */
static void
fetch_keys (HWND dlg, listview_ctrl_t lv, keyserver_ctx *ksc)
{
    char keyid[32];
    char *fpr;
    int n, rc = 0;

    n = listview_count_items (lv, 1);
    if (!n) {
	msg_box (dlg, _("Please select a key."), _("Keyserver Search"), MB_ERR);
	return;
    }

    for (n=0; n < listview_count_items (lv, 0); n++) {
	if (!listview_get_item_state (lv, n))
	    continue;

	listview_get_item_text (lv, n, KSS_COL_KEYID, keyid, DIM (keyid)-1);
	rc = hkp_recv_key2 (dlg, ksc->name, ksc->port, keyid, 0, &fpr);
	if (!rc && fpr != NULL) {
	    keycache_update (0, fpr);
	    free_if_alloc (fpr);
	}
    }
}


/* Dialog box procedure for keyserver searching. */
BOOL CALLBACK
hkpsearch_dlg_proc (HWND dlg, UINT msg, WPARAM wparam, LPARAM lparam)
{
    static listview_ctrl_t lv = NULL;
    static keyserver_ctx *ksc;
    
    switch (msg) {
    case WM_INITDIALOG:
        ksc = (keyserver_ctx *)lparam;
	if (!ksc)
	    BUG (0);
	lv = on_init_dialog (dlg, msg, wparam, lparam);
	if (search_hkp_keys (dlg, ksc, lv)) {
	    EndDialog (dlg, FALSE);
	    return TRUE;
	}
        return TRUE;
        
    case WM_DESTROY:
        if (lv) {
            listview_release (lv);
            lv = NULL;
        }
        return FALSE;
        
    case WM_COMMAND:
        switch (LOWORD (wparam)) {
        case IDOK:
            EndDialog (dlg, TRUE);
            return TRUE;
            
        case IDCANCEL:
            EndDialog (dlg, FALSE);
            return TRUE;
            
        case IDC_HKPSEARCH_RECV:
	    fetch_keys (dlg, lv, ksc);
            return TRUE;
        }
        break;
    }
    
    return FALSE;
}
